// Author 		: F. de la Parra
// Date			: February, 2019
// Description 	: Each method of class FileIODemo exercises file I/O
//				  (text and binary files)and associated exception handling.  
// Version		: 1.0
 
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Scanner;
import java.util.InputMismatchException;
import java.util.NoSuchElementException;
import java.util.Random;

public class FileIODemo {
	
	// Method "main" drives topic selection
	
	public static void main(String[] args) {
		
		// Create a 'demo' object.
		FileIODemo demoObject = new FileIODemo();
		
		// Create Scanner object 'menuSelect' for input from keyboard. 
		Scanner menuSelect = new Scanner(System.in);
		
		// Initialize menuItem selection.
		int menuItem = 0;
		
		// Display menu and wait for a menu item selection
		while (menuItem != 9) {
			// Display menu
			System.out.println("\n***FileIODemo Menu ***");
			System.out.println("1 - File Output");
			System.out.println("2 - File Input");
			System.out.println("3 - Write Binary File");
			System.out.println("4 - Read Binary File");
			System.out.println("9 - Exit");
			System.out.println();
			System.out.println("Enter demo number:");
			
			// Input a menu selection.
			menuItem = menuSelect.nextInt();
			
			// Run demo corresponding to selected menu item.
			switch (menuItem) {
				case 1 : {
					System.out.println("\n*** Running 'file output' demo ***\n");
					
					// Execute file output demo.
					demoObject.fileOutputDemo();
					break;
				}
				
				case 2 : {
					System.out.println("\n*** Running 'file input' demo ***\n");
					
					// Execute file input demo.
					demoObject.fileInputDemo();
					break;
				}
				
				case 3 : {
					System.out.println("\n*** Running write binary file demo ***\n");
					
					// Execute file input demo.
					demoObject.writeBinaryFileDemo();
					break;
				}
				
				case 4 : {
					System.out.println("\n*** Running read binary file demo ***\n");
					
					// Execute file input demo.
					demoObject.readBinaryFileDemo();
					break;
				}
				
			 // case n :     More upcoming demos ...
				
				default : {
					System.out.println("Either 9 or invalid selection entered");
					break;
				}
			}
			System.out.println("\n*** Demo completed ***\n");
		}
		
		System.out.println("\n*** Program terminated ***\n");
	}
	
	// Method fileOutputDemo exercises writing data to a text file.
	
	public void fileOutputDemo() {
		
		// Declare a data stream object to send data to a text file.
		PrintWriter toFileStream = null;
		
		// Open the associated file for 'WRITE'. Erase any existing content in test.txt.
		// Check for file I/O exceptions.
		try {
			toFileStream = new PrintWriter(new FileOutputStream("test.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'test.txt'. FileIODemo aborted!");
			System.exit(0);
		}

		// WRITE something to the file. For example, in a for loop with 10 steps.
		int i, r1;
		double r2, r3;
		
		for (i = 1; i <= 10; i++) {
			
			// Generate 3 random numbers. One integer and two floating point numbers.
			Random generator = new Random(System.currentTimeMillis());
			r1 = generator.nextInt();
			r2 = (double)generator.nextInt();
			r3 = (double)generator.nextInt();
			
			// Write two lines to file test.txt at each step of the for loop.
			
			// The first line contains just a stream of unformatted characters.
			toFileStream.println("The content of line " + (2*i-1) + " is just a String of characters.");
			
			// The second line contains formatted and unformatted data.
			toFileStream.printf("Line %d contains 3 random numbers:%d,%6.2f,%e\n",
								2*i, r1, r2, r3);
		}
		
		// Closing message to the user.
		System.out.println("Inserted " + 2*(i - 1) + " lines in the file 'test.txt'");
		
		// Close the file to release system resources associated with it.
		toFileStream.close();	
	}
	
	// Method fileInputDemo exercises reading from a text file.

	public void fileInputDemo() {
		
		// Declare a data stream object to read data from a text file.
		Scanner fromFileStream = null;

		// Open the file. Check for file I/O exceptions.
		try {
			fromFileStream = new Scanner(new FileInputStream("test.txt"));
		}
		catch (FileNotFoundException e) {	
			System.out.println("Error opening the file 'test.txt'. FileIODemo aborted!");
			System.exit(0);
		}
		
		// Create object to read input from keyboard. 
		Scanner keyboard = new Scanner(System.in);
		
		// Dump input file to screen (D) or selectively parse its lines (P).
		String option = null;
		System.out.print("(D)ump unchanged to screen or (P)arse input file (D/P)?: ");
		
		// Get input from the keybord and convert it to uppercase.
		option = keyboard.next().toUpperCase();
		
		while (!(option.equals("D") || option.equals("P"))) {
			System.out.println("Invalid option! Must be D or P!");
			System.out.print("Dump to Screen or Parse Input File (D/P): ");
			option = keyboard.next();
		}
		
		// Process file 'test.txt', generated by method 'fileOutputDemo'.
		int lineNumber = 1;
		
		if (option.equals("D")) {
			
			// 'READ' entire file, line-by-line, until END OF FILE. Dump it on the screen.
			String inputLine;
			
			// Message to the user about parsing of input file.
			System.out.println("\nDUMPING TO THE SCREEN FILE 'test.txt'\n");
			
			while (fromFileStream.hasNextLine()) {
				inputLine = fromFileStream.nextLine();
				System.out.println("Line " + lineNumber + ": " + inputLine);
				lineNumber += 1;
			}
		}
		else {
			// Selectively PARSE lines of the input. Display formatted output on the screen.
			
			// Types of tokens that can be found in file 'test.txt'.
			String inputToken = null;
			int intToken;
			double doubleToken1;
			double doubleToken2;
			
			// Pattern that is found in the unformatted lines of 'test.txt'.
			String stringPattern = "String";

			// Tokens can be delimited by ":", "," or "\n".
			fromFileStream.useDelimiter("[:,\n]");
			
			// File processing control flag.
			boolean done = false;
			
			// 'READ' entire file, line-by-line, until END OF FILE. Display values on screen.
			
			// Read next token. If the input file 'test.txt' is empty, abort FileIODemo.
			try {
				inputToken = fromFileStream.next();
			}
			catch (NoSuchElementException e) {	
				System.out.println("Empty file 'test.txt'! FileIODemo terminated!");
				System.exit(0);
			}
			
			// Message to the user about parsing of input file.
			System.out.println("\nPARSING OF LINES IN 'test.txt'\n");
			
			while (!done) {
				if (inputToken.contains("String")) {
					// Dump odd lines to the screen. They just contain a string.
					System.out.println("Odd line: " + inputToken + "\n");
					fromFileStream.nextLine();
				}
				else {
					System.out.println("Even line: " + inputToken);
					intToken = fromFileStream.nextInt();
					System.out.println("Integer: " + intToken);
					doubleToken1 = fromFileStream.nextDouble();
					System.out.println("Double: " + doubleToken1);
					doubleToken2 = fromFileStream.nextDouble();
					System.out.println("Double: " + doubleToken2 + "\n");
				}
				
				// Read the first token in the next line.
				// Ignore empty lines, just terminated with "\n".
				if (fromFileStream.hasNextLine())
					try {
						inputToken = fromFileStream.next();
					}
					catch (NoSuchElementException e) {	
						System.out.println("Empty line in 'test.txt'. Line skipped!");
						done = true;
					}
				else
					done = true;
			}
		}
		
		// Close the file to release system resources associated with it.
		fromFileStream.close();
	}
	
	// Method writeBinaryFileDemo exercises writing to a binary file.

	public void writeBinaryFileDemo() {	
	
		// Declare a binary stream object to write data to a binary file.
		ObjectOutputStream toFileBinaryStream = null;
		
		// Open the file for writing. Check for file I/O exceptions.
		try {
			toFileBinaryStream = new ObjectOutputStream(new FileOutputStream("data.obj"));
		}
		catch (IOException e) {	
			System.out.println("Error opening the file 'data.obj'. FileIODemo aborted!");
			System.exit(0);
		}
		
		// Create object to read input from keyboard. 
		Scanner keyboard = new Scanner(System.in);
		
		// Initialize variables to write to file 'data.obj'
		int intNum = -8;
		boolean flag = true;
		float floatNum = 54.89f;
		double doubleNum = 4.5;
		char charVar = 'a';
		String stringVar = "Hello";
		
		// Create Student object whose state can be written to a binary stream
		Student John = new Student(12345678, "John", "Doe", 1);
		
		try {
			toFileBinaryStream.writeInt(intNum);
			toFileBinaryStream.writeBoolean(flag);
			toFileBinaryStream.writeFloat(floatNum);
			toFileBinaryStream.writeDouble(doubleNum);
			toFileBinaryStream.writeChar(charVar);
			toFileBinaryStream.writeUTF(stringVar);
		}
		catch (IOException e) {
			System.out.println("Problems writing to 'data.obj' file. FileIODemo aborted!");
		}
		
		// Write the state of an object whose class definition implements the
		// Serializable interface.
		try {
			toFileBinaryStream.writeObject(John);
		}
		catch (IOException e) {
			System.out.println("Student object's state written to binary file");
		}
		
		// Close the binary file stream.
		try {
			toFileBinaryStream.close();
		}
		catch (IOException e) {
			System.out.println("Problems closing 'data.obj' file. FileIODemo aborted!");
		}
	}
	
	// Method readBinaryFileDemo exercises reading from a binary file.

	public void readBinaryFileDemo() {	
		
		// Declare a binary stream object to read data from a binary file.
		ObjectInputStream fromFileBinaryStream = null;
		
		// Open the file for reading. Check for file I/O exceptions.
		try {
			fromFileBinaryStream = new ObjectInputStream(new FileInputStream("data.obj"));
		}
		catch (IOException e) {	
			System.out.println("Error opening the file 'data.obj'. FileIODemo aborted!");
			System.exit(0);
		}
		

		// Read values from a binary file back to variables, in the
		// same order they were written out.
		try {

			int readIntNum = fromFileBinaryStream.readInt();
			boolean readFlag = fromFileBinaryStream.readBoolean();
			float readFloatNum = fromFileBinaryStream.readFloat();
			double readDoubleNum = fromFileBinaryStream.readDouble();
			char readCharVar = fromFileBinaryStream.readChar();
			String readStringVar = fromFileBinaryStream.readUTF();

			// Display values on screen.
			System.out.println("Integer: " + readIntNum);
			System.out.println("Boolean: " + readFlag);
			System.out.println("Float: " + readFloatNum);
			System.out.println("DoubleInteger: " + readDoubleNum);
			System.out.println("Character: " + readCharVar);
			System.out.println("String: " + readStringVar);
		}
		catch (IOException e) {
			System.out.println("Problems reading from 'data.obj' file. FileIODemo aborted!");
			System.exit(0);
		}
		
		// Read and display a student object from 'data.obj'. Reference it with Student
		// variable 'John'.
		try {
			// Read the object's state back to a student object referenced by John.
			Student John = (Student)fromFileBinaryStream.readObject();
			
			// Display class type and reference to object read from binary file 'data.obj'.
			System.out.println("\nObject referenced by 'John': " + John);
			
			// Display state of object read from binary file 'data.obj'.
			System.out.println("Student number: " + John.getStudentNumber());
			System.out.println("Student first name: " + John.getFirstName());
			System.out.println("Student last name: " + John.getLastName());
		}
		catch (ClassNotFoundException e) {
			// No class definition for object read from binary file 'data.obj'.
			System.out.println("Class not found for object from 'data.obj'. FileIODemo aborted!");
			System.exit(0);
		}
		catch (IOException e) {
			System.out.println("Unable to read object from 'data.obj'. FileIODemo aborted!");
			System.exit(0);
		}
		
		// Close the binary file.
		try {
			fromFileBinaryStream.close();
		}
		catch (IOException e) {
			System.out.println("Problems closing 'data.obj' file. FileIODemo aborted!");
		}
	}
}