import java.util.InputMismatchException;
import java.io.Serializable;
// Author: F. de la Parra
// Date  : February, 2019
// Description: The Student class serves as a "template" to create
// objects of type Student.  

public class Student implements Serializable {
	// Student's basic personal data
	private int studentNumber;
	private String firstName;
	private String lastName;
	private int programYear;
	
	// Student's additional academic data
	private int startYear;
	private boolean doubleMajor;
	private double gpa;
	
	// Constructors.
	
	public Student(int studentNumber, String firstName, String lastName, int programYear,
			int startYear, boolean doubleMajor, double gpa) throws InputMismatchException {
		
		try {
			setStudentNumber(studentNumber);
		}
		catch (InputMismatchException e) {
			System.out.println("Invalid student number! Program aborted!");
			System.exit(0);
		}
		
		this.firstName = firstName;
		this.lastName = lastName;
		this.programYear = programYear;
		this.startYear = startYear;
		this.doubleMajor = doubleMajor;
		this.gpa = gpa;
	}
	
	public Student(int studentNumber, String firstName, String lastName, int programYear)
			throws InputMismatchException {
		
		try {
			setStudentNumber(studentNumber);
		}
		catch (InputMismatchException e) {
			System.out.println("Invalid student number! Program aborted!");
			System.exit(0);
		}
		
		this.firstName = firstName;
		this.lastName = lastName;
		this.programYear = programYear;
	}
	
	// Student mutator methods
	
	public boolean setStudentNumber(int studentNumber) throws InputMismatchException {
		
		boolean status = false;
		if (studentNumber >= 10000000 && studentNumber < 99999999)
			this.studentNumber = studentNumber;
		else
			throw new InputMismatchException("Invalid student number");
		return true;
	}

	public void setStartYear(int startYear) {
		this.startYear = startYear;
	}

	public void setDoubleMajor(boolean doubleMajor) {
		this.doubleMajor = doubleMajor;
	}

	public void setGpa(double gpa) {
		this.gpa = gpa;
	}
	
	// Student accessor methods
	
	public int getStudentNumber() {
		return this.studentNumber;
	}
	
	public String getFirstName() {
		return this.firstName;
	}

	public String getLastName() {
		return this.lastName;
	}
	
	public int getStartYear() {
		return this.startYear;
	}
	
	public boolean getDoubleMajor() {
		return this.doubleMajor;
	}
	
	public double getGpa() {
		return this.gpa;
	}
	
	// Miscellaneous methods
	
	public String description() {
		String allText = "Student " + this.firstName + " " + 
		                 this.lastName + " is in programYear " + this.programYear;
		return allText;
	}
}